#!/usr/bin/perl -w

# Notes:
#
# Use of $', $`, $&, $1, $2... cause perl to create these on every match
# (thereby copying the entire string)
# use /^.{$pos}/ to mark position $pos in string? and /.{$pos}$/ from end.
#   is this faster than \G?

# program to test ad finding algorithms

use Time::HiRes;
use HTML::TokeParser;
use HTML::Parser;

open(FILE, "<$ARGV[0]") || die "need filename\n";
$doc = join("", <FILE>);
close FILE;
$whitespace = " |\s|\013|\015|\n|\r|\f|\t";
$basicad  = "(?:<(?:$whitespace)*i?layer(?:$whitespace)[^>]*>.*?<(?:$whitespace)*/i?layer[^>]*>|<(?:$whitespace)*a(?:$whitespace)[^>]*href[^>]*>.*?<(?:$whitespace)*\/a(?:$whitespace)*>|<(?:$whitespace)*i?frame(?:$whitespace)[^>]*>.*?<(?:$whitespace)*/i?frame[^>]*>)";
#$basicad  = "<(?:$whitespace)*i?layer(?:$whitespace)[^>]*>.*?<(?:$whitespace)*/i?layer[^>]*>";
# <!--.*?--> isn't working properly.  It picks up <!--..-->blah<!--..--> as a single comment.
$dontcare = "$whitespace|&nbsp;|<[bh]r[^>]*>|<p[^>]*?>|<\!--(.*?)-->";
$dontcarescript = "<(?:$whitespace)*script[^>]*?>(.*?)<(?:$whitespace)*/script[^>]*?>";
my($dontcareilayer) = "<(?:$whitespace)*ilayer[^>]*?>(.*?)<(?:$whitespace)*/ilayer[^>]*?>";
my($dontcarelayer) = "<(?:$whitespace)*layer[^>]*?>(.*?)<(?:$whitespace)*/layer[^>]*?>";
#$adregexp = "/banners?/|/ads?/|adverts?|adserver|linkexchange\.com|flycast\.com|clicktrade\.com|doubleclick\.net|avenuea\.com|blockstackers\.com|valueclick\.com|http://ad[a-z\.]+|enliven";
$adregexp = "/banners?/|/ads?/|adserver|/adclick/|adverts?|blipverts?|promo|sponsor|http://(?:ad\.|ads\.|adserver\.)|linkexchange\.com|flycast\.com|clicktrade\.com|doubleclick\.net|avenuea\.com|blockstackers\.com|mediaplex\.com|ads[0-9]+\.focalink\.com|valueclick\.com|yimg\.com|imgis\.com|bfast\.com|admaximize\.com|eads\.com|http://192\.41\.13|datais\.com|commission-junction\.com|pennyweb\.com|store.astronauts.org|http://www.space.com/entertainment/spacequiz/index.php3|preferences\.com";
$iterations = $ARGV[1] || 1;
$printstuff = $iterations==1?1:0;

for($i=1;$i<=1;$i++) {
  $begin = Time::HiRes::time;
  ($buser, $bsys) = times;
  for($j=0;$j<$iterations;$j++) {
    &{"finder" . $i }($doc);
  }
  $end = Time::HiRes::time;
  ($euser, $esys) = times;
  print "finder$i took ", ($end-$begin)/$j, " clock, ", 
                         ($euser-$buser)/$j, " user, ",
                         ($esys-$bsys)/$j, " system seconds.\n";
}

sub debug {
  if($printstuff) {
    print @_;
  }
}

sub finder1 {
  my($doc) = @_;
  my($numbasicads) = 0;
  my($numads) = 0;
  my($ad);

  pos($doc) = 0;
BASICAD:
  while($doc =~ /\G.*?($basicad)/sig) {
    $ad = $1;
    $numbasicads++;
    print '-' x 80 . "\n";
    my($ad1) = $ad;
    $ad1 =~ tr/\015/\n/;
    print $ad1 . "\n";
    next BASICAD unless($ad =~ /(img|src)/si);
    print "has an img...\n";
    next BASICAD unless($ad =~ $adregexp);
    print "matched adregexp...\n";
    $endpos     = pos($doc);
    $matchpos   = pos($doc) - length $ad;
    $numads++;
# suck in surrounding useless tags
    my($tag);
    my($tmpmatchpos) = $matchpos;
    my($tmpendpos) = $endpos;
ENCLOSINGTAG:
    while(1) {
      pos($doc) = $endpos;
      print "Examining: '", substr($doc, $endpos, 10), "'\n";
      last ENCLOSINGTAG unless $doc =~ /\G(?:$dontcare)*<\s*\/([a-z]+)[^>]*>(?:$dontcare)*/sig;
      $tag = $2; # note 1 parenthesis in $dontcare
      $comment1 = $1;
      $comment2 = $3;
      last ENCLOSINGTAG if(defined $comment1 && $comment1 =~ /-->|<!--/);
      last ENCLOSINGTAG if(defined $comment2 && $comment2 =~ /-->|<!--/);
      print "Next trailing tag is $tag\n";
      $tmpendpos = pos($doc);
      $tmpmatchpos = $matchpos;
      do {
        do {
          do {
            $tmpmatchpos--;
          } while((substr($doc, $tmpmatchpos, 1) ne '<') && $tmpmatchpos>0);
        } while(substr($doc, $tmpmatchpos, $matchpos-$tmpmatchpos) =~ /^(?:$dontcare)*$/si);
        pos($doc) = $tmpmatchpos;
      } while($doc !~ /\G(<\s*\/?[a-z]*[^>]*?>)/sig); 
# find previous tag
      pos($doc) = $tmpmatchpos;
      last ENCLOSINGTAG unless $doc =~ /\G<\s*$tag[^>]*>/sig;
      $matchpos = $tmpmatchpos;
      $endpos = $tmpendpos;
      pos($doc) = $tmpendpos;
      if($tag =~ /noscript|nolayer/si) { 
        my($dontcareopp, $dontcaretag);
        if($tag =~ /noscript/si) {
          $dontcareopp = $dontcarescript;
          $dontcaretag = "script";
        }
        if($tag =~ /nolayer/si) {
          $dontcareopp = "(?:" . $dontcarelayer . "|" . $dontcareilayer . ")";
          $dontcaretag = "(?:ilayer|layer)";
        }
        if($doc =~ /\G(?:$dontcare)*?$dontcarescript(?:$dontcare)*/sig) {
          $endpos = pos($doc);
        } else {
          do {
            do {
              $tmpmatchpos--;
            } while((substr($doc, $tmpmatchpos, 1) ne '<') && $tmpmatchpos>0);
          } while(substr($doc, $tmpmatchpos, $matchpos-$tmpmatchpos) =~ /^(?:$dontcare)*$/si);
          pos($doc) = $tmpmatchpos;
          while($doc =~ /\G<\s*\/$dontcaretag[^>]*?>/sig) { # suck in <script></script>
            do {
              do {
                $tmpmatchpos--;
              } while((substr($doc, $tmpmatchpos, 1) ne '<') && $tmpmatchpos>0);
            } while(substr($doc, $tmpmatchpos, $matchpos-$tmpmatchpos) !~ /^<\s*$dontcaretag[^>]*>/si 
                && $tmpmatchpos>0);
            if($tmpmatchpos > 0) {
              $matchpos = $tmpmatchpos;
            }
            do { # rewind to previous tag, check if it is </script>
              do {
                $tmpmatchpos--;
              } while((substr($doc, $tmpmatchpos, 1) ne '<') && $tmpmatchpos>0);
            } while(substr($doc, $tmpmatchpos, $matchpos-$tmpmatchpos) =~ /^(?:$dontcare)*$/si);
            pos($doc) = $tmpmatchpos;
          }
        }
      }
    }
    debug(substr($doc, $matchpos, $endpos-$matchpos) . "\n" . '-' x 80 . "\n");
    substr($doc, $matchpos, $endpos-$matchpos, "");
    pos($doc) = $matchpos;
  }
  debug("Found $numbasicads basic ads, $numads real ads\n");
}

sub finder2 {
  my($doc) = @_;
  my($numbasicads) = 0;
  my($numads) = 0;
  my($ad);
  my($tag);
  my($matchpos) = 0;
  my($endpos) = 0;
  my($tmpmatchpos) = 0;

  pos($doc) = 0;
BASICAD:
  while($doc =~ /\G.*?($basicad)/sig) {
    $numbasicads++;
    $ad = $1;
    next BASICAD unless $ad =~ /(img|src)/si;
    next BASICAD unless $ad =~ $adregexp;
    $numads++; 
    $matchpos = pos($doc) - length $ad;
    $tmpmatchpos = $matchpos;
    $endpos = pos($doc);
    pos($doc) = $matchpos;
    while( # Welcome to the most hideous piece of code I have ever written.
           (
             eval {
               do {
                 $tmpmatchpos--;
                 while(substr($doc, $tmpmatchpos, 1) ne '<' && $tmpmatchpos>0) {
                   $tmpmatchpos--; 
                 }
               } while(substr($doc, $tmpmatchpos, $matchpos-$tmpmatchpos) =~ /^($dontcare)+$/si
                       && $tmpmatchpos>0);
               pos($doc) = $tmpmatchpos;
               1;
             } || 1
           ) && $doc =~ /\G($dontcare)*<([a-z]+)[^>]*>($dontcare)*/sig 
             && ($tag = $2) && (pos($doc) = $endpos) 
             && $doc =~ /\G($dontcare)*<\/$tag[^>]*>($dontcare)*/sig
         ) {
#        debug("sucking in <$tag> to ad.\n");
      $matchpos = $tmpmatchpos;
      $endpos = pos($doc);
    }
    pos($doc) = $endpos;
#      debug(substr($doc, $matchpos, $endpos-$matchpos) . "\n");
  }
  debug("Found $numbasicads basic ads, $numads real ads\n");
}

sub finder3 {
  my($doc) = @_;
  my($p) = HTML::TokeParser->new(\$doc);
  my($numbasicads) = 0;
  my($numads) = 0;

  while(my $token = $p->get_tag("a")) {
    $numbasicads++;
    my($url) = $token->[1]{href} || "";
    my($text) = $p->get_text("/a");
    if($url =~ $adregexp || $text =~ $adregexp) {
      $numads++;
    }
  }
  debug("Found $numbasicads basic ads, $numads real ads\n");
}

sub finder4 {
  my($doc) = @_;
  my($lasta) = 0; # pos of last <a href...>
  my($numbasicads) = 0;
  my($numads) = 0;
  my($needenclosers) = 0;
  my($insidea) = 0;
  my(@prevtags); # list of previous tags.
  my($matchpos, $endpos);
  my($i);

  my($start) = sub {
    my($tag, $offset, $len, $is_cdata) = @_;
    if($tag !~ /^(?:[hb]r|p|img|li|input|font|area)$/) { # tags that are not balanced.
      if(!$insidea) {
        push @prevtags, [$tag, $offset, $len]; # add this tag to list
      }
      if($tag =~ /^([hb]r|p)$/ && $needenclosers) {
        $endpos = $offset+$len;
      } elsif($needenclosers) {
debug(substr($doc, $matchpos, $endpos-$matchpos) . "\n\n");
        $needenclosers = 0;
      }
    }
    if($tag eq 'a' && !$is_cdata && !$insidea) {
      $numbasicads++;
      $lasta = $offset;
      $matchpos = $offset;
      $insidea = 1;
    }
  };
  my($end) = sub {
    my($tag, $offset, $len, $is_cdata) = @_;
    if(!$insidea) {
PREVTAGS:
      for($i=$#prevtags;$i>0;$i--) {
        if($prevtags[$i] && $tag eq $prevtags[$i]->[0]) {
          $prevtag = $prevtags[$i];
          if($i == $#prevtags) {
            pop @prevtags;
          } elsif($i>0 && $i<$#prevtags) {
            @prevtags = (@prevtags[0..$i-1], @prevtags[$i+1..$#prevtags]);
          } elsif($i==0) {
            @prevtags = (@prevtags[1..$#prevtags]);
          }
          last PREVTAGS;
        }
      }
    }
    if($tag eq 'a' && !$is_cdata) {
      my($ad) = substr($doc, $lasta, $offset+$len-$lasta);
      if($ad =~ /(img|src)/si && $ad =~ $adregexp) {
        $numads++;
        $needenclosers=1;
        $endpos = $offset+$len;
      }
      $insidea = 0;
    } elsif ($needenclosers) {
      if($prevtag) {
        if(substr($doc, $prevtag->[1]+$prevtag->[2], $matchpos-$prevtag->[1]-$prevtag->[2]) =~ /^(?:$dontcare)*$/si) {
          $one = $1;
          $two = $2;
          if((!$one || $one !~ /(?:<!--|-->)/si) && (!$two || $two !~ /(?:<\s*\/?script[^>]*>)/si)) {
#debug("sucking in <$tag> to ad.\n");
            $matchpos = $prevtag->[1];
            $endpos = $offset+$len;
          }
        }
      } elsif ($tag =~ /[hb]r|p/) {
        $endpos = $offset+$len;
      }
    }
  };
  my($p) = HTML::Parser->new( api_version => 3);
  $p->handler(start => $start, 'tagname, offset, length, is_cdata');
  $p->handler(end => $end, 'tagname, offset, length, is_cdata');
  $p->parse($doc);
  $p->eof;
  debug("Found $numbasicads basic ads, $numads real ads\n");
}
